@@ -16,8 +16,9 @@
 #include <string.h>
 #include <sys/types.h>
 
+#include <gio/gio.h>
+
 #include "soup-address.h"
-#include "soup-dns.h"
 #include "soup-enum-types.h"
 #include "soup-marshal.h"
 #include "soup-misc.h"
@@ -60,9 +61,6 @@ typedef struct {
 
 	char *name, *physical;
 	guint port;
-
-	SoupDNSLookup *lookup;
-	guint timeout_id;
 } SoupAddressPrivate;
 #define SOUP_ADDRESS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_ADDRESS, SoupAddressPrivate))
 
@@ -151,11 +149,6 @@ finalize (GObject *object)
 	if (priv->physical)
 		g_free (priv->physical);
 
-	if (priv->lookup)
-		soup_dns_lookup_free (priv->lookup);
-	if (priv->timeout_id)
-		g_source_remove (priv->timeout_id);
-
 	G_OBJECT_CLASS (soup_address_parent_class)->finalize (object);
 }
 
@@ -164,8 +157,6 @@ soup_address_class_init (SoupAddressClass *address_class)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (address_class);
 
-	soup_dns_init ();
-
 	g_type_class_add_private (address_class, sizeof (SoupAddressPrivate));
 
 	/* virtual method override */
@@ -175,6 +166,12 @@ soup_address_class_init (SoupAddressClass *address_class)
 	object_class->get_property = get_property;
 
 	/* properties */
+	/**
+	 * SOUP_ADDRESS_NAME:
+	 *
+	 * Alias for the #SoupAddress:name property. (The hostname for
+	 * this address.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_NAME,
 		g_param_spec_string (SOUP_ADDRESS_NAME,
@@ -182,6 +179,12 @@ soup_address_class_init (SoupAddressClass *address_class)
 				     "Hostname for this address",
 				     NULL,
 				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_ADDRESS_FAMILY:
+	 *
+	 * Alias for the #SoupAddress:family property. (The
+	 * #SoupAddressFamily for this address.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_FAMILY,
 		g_param_spec_enum (SOUP_ADDRESS_FAMILY,
@@ -190,6 +193,12 @@ soup_address_class_init (SoupAddressClass *address_class)
 				   SOUP_TYPE_ADDRESS_FAMILY,
 				   SOUP_ADDRESS_FAMILY_INVALID,
 				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_ADDRESS_PORT:
+	 *
+	 * An alias for the #SoupAddress:port property. (The port for
+	 * this address.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_PORT,
 		g_param_spec_int (SOUP_ADDRESS_PORT,
@@ -197,6 +206,12 @@ soup_address_class_init (SoupAddressClass *address_class)
 				  "Port for this address",
 				  -1, 65535, -1,
 				  G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	/**
+	 * SOUP_ADDRESS_PHYSICAL:
+	 *
+	 * An alias for the #SoupAddress:physical property. (The
+	 * stringified IP address for this address.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_PHYSICAL,
 		g_param_spec_string (SOUP_ADDRESS_PHYSICAL,
@@ -204,6 +219,12 @@ soup_address_class_init (SoupAddressClass *address_class)
 				     "IP address for this address",
 				     NULL,
 				     G_PARAM_READABLE));
+	/**
+	 * SOUP_ADDRESS_SOCKADDR:
+	 *
+	 * An alias for the #SoupAddress:sockaddr property. (A pointer
+	 * to the struct sockaddr for this address.)
+	 **/
 	g_object_class_install_property (
 		object_class, PROP_SOCKADDR,
 		g_param_spec_pointer (SOUP_ADDRESS_SOCKADDR,
@@ -235,12 +256,7 @@ constructor (GType                  type,
 		return NULL;
 	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
 
-	if (priv->name) {
-		if (!priv->sockaddr)
-			priv->lookup = soup_dns_lookup_name (priv->name);
-	} else if (priv->sockaddr)
-		priv->lookup = soup_dns_lookup_address (priv->sockaddr);
-	else {
+	if (!priv->name && !priv->sockaddr) {
 		g_object_unref (addr);
 		return NULL;
 	}
@@ -461,10 +477,23 @@ soup_address_get_sockaddr (SoupAddress *addr, int *len)
 
 	if (priv->sockaddr && len)
 		*len = SOUP_ADDRESS_FAMILY_SOCKADDR_SIZE (SOUP_ADDRESS_GET_FAMILY (priv));
-
 	return priv->sockaddr;
 }
 
+static GInetAddress *
+soup_address_make_inet_address (SoupAddress *addr)
+{
+	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	GSocketAddress *gsa;
+	GInetAddress *gia;
+
+	gsa = g_socket_address_new_from_native (priv->sockaddr,
+						SOUP_ADDRESS_FAMILY_SOCKADDR_SIZE (SOUP_ADDRESS_GET_FAMILY (priv)));
+	gia = g_inet_socket_address_get_address ((GInetSocketAddress *)gsa);
+	g_object_unref (gsa);
+	return gia;
+}
+
 /**
  * soup_address_get_physical:
  * @addr: a #SoupAddress
@@ -485,8 +514,13 @@ soup_address_get_physical (SoupAddress *addr)
 	if (!priv->sockaddr)
 		return NULL;
 
-	if (!priv->physical)
-		priv->physical = soup_dns_ntop (priv->sockaddr);
+	if (!priv->physical) {
+		GInetAddress *gia;
+
+		gia = soup_address_make_inet_address (addr);
+		priv->physical = g_inet_address_to_string (gia);
+		g_object_unref (gia);
+	}
 
 	return priv->physical;
 }
@@ -508,46 +542,122 @@ soup_address_get_port (SoupAddress *addr)
 }
 
 
-static void
-update_address (SoupAddress *addr, SoupDNSLookup *lookup)
+static guint
+update_addrs (SoupAddress *addr, GList *addrs, GError *error)
+{
+	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	GInetAddress *gia;
+	GSocketAddress *gsa;
+	gsize len;
+
+	if (error) {
+		if (error->domain == G_IO_ERROR &&
+		    error->code == G_IO_ERROR_CANCELLED)
+			return SOUP_STATUS_CANCELLED;
+		else
+			return SOUP_STATUS_CANT_RESOLVE;
+	} else if (!addrs)
+		return SOUP_STATUS_CANT_RESOLVE;
+	else if (priv->sockaddr)
+		return SOUP_STATUS_OK;
+
+	gia = addrs->data;
+	gsa = g_inet_socket_address_new (gia, priv->port);
+
+	len = g_socket_address_get_native_size (gsa);
+	priv->sockaddr = g_malloc (len);
+	g_socket_address_to_native (gsa, priv->sockaddr, len, NULL);
+
+	return SOUP_STATUS_OK;
+}
+
+static guint
+update_name (SoupAddress *addr, const char *name, GError *error)
 {
 	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
 
-	if (!priv->name)
-		priv->name = soup_dns_lookup_get_hostname (lookup);
+	if (error) {
+		if (error->domain == G_IO_ERROR &&
+		    error->code == G_IO_ERROR_CANCELLED)
+			return SOUP_STATUS_CANCELLED;
+		else
+			return SOUP_STATUS_CANT_RESOLVE;
+	} else if (!name)
+		return SOUP_STATUS_CANT_RESOLVE;
+	else if (priv->name)
+		return SOUP_STATUS_OK;
 
-	if (!priv->sockaddr) {
-		priv->sockaddr = soup_dns_lookup_get_address (lookup);
-		SOUP_ADDRESS_SET_PORT (priv, htons (priv->port));
-	}
+	priv->name = g_strdup (name);
+	return SOUP_STATUS_OK;
 }
 
 typedef struct {
 	SoupAddress         *addr;
 	SoupAddressCallback  callback;
 	gpointer             callback_data;
+	gboolean             lookup_name;
+
+	GMainContext        *async_context;
+	GCancellable        *cancellable;
+	guint                status;
 } SoupAddressResolveAsyncData;
 
 static void
-lookup_resolved (SoupDNSLookup *lookup, guint status, gpointer user_data)
+complete_resolve_async (SoupAddressResolveAsyncData *res_data)
 {
-	SoupAddressResolveAsyncData *res_data = user_data;
-	SoupAddress *addr;
-	SoupAddressCallback callback;
-	gpointer callback_data;
-
-	addr = res_data->addr;
-	callback = res_data->callback;
-	callback_data = res_data->callback_data;
-	g_free (res_data);
-
-	if (status == SOUP_STATUS_OK)
-		update_address (addr, lookup);
+	SoupAddress *addr = res_data->addr;
+	SoupAddressCallback callback = res_data->callback;
+	gpointer callback_data = res_data->callback_data;
 
 	if (callback)
-		callback (addr, status, callback_data);
+		callback (addr, res_data->status, callback_data);
+
+	if (res_data->async_context)
+		g_main_context_unref (res_data->async_context);
+	if (res_data->cancellable)
+		g_object_unref (res_data->cancellable);
 
 	g_object_unref (addr);
+	g_slice_free (SoupAddressResolveAsyncData, res_data);
+}
+
+static void
+lookup_resolved (GObject *source, GAsyncResult *result, gpointer user_data)
+{
+	GResolver *resolver = G_RESOLVER (source);
+	SoupAddressResolveAsyncData *res_data = user_data;
+	SoupAddress *addr = res_data->addr;
+	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	GError *error = NULL;
+
+	if (!priv->sockaddr) {
+		GList *addrs;
+
+		addrs = g_resolver_lookup_by_name_finish (resolver, result,
+							  &error);
+		res_data->status = update_addrs (addr, addrs, error);
+		g_resolver_free_addresses (addrs);
+	} else if (!priv->name) {
+		char *name;
+
+		name = g_resolver_lookup_by_address_finish (resolver, result,
+							    &error);
+		res_data->status = update_name (addr, name, error);
+		g_free (name);
+	} else
+		res_data->status = SOUP_STATUS_OK;
+
+	if (error)
+		g_error_free (error);
+
+	complete_resolve_async (res_data);
+}
+
+static gboolean
+idle_complete_resolve (gpointer res_data)
+{
+	complete_resolve_async (res_data);
+	return FALSE;
 }
 
 /**
@@ -585,18 +695,48 @@ soup_address_resolve_async (SoupAddress *addr, GMainContext *async_context,
 {
 	SoupAddressPrivate *priv;
 	SoupAddressResolveAsyncData *res_data;
+	GResolver *resolver;
 
 	g_return_if_fail (SOUP_IS_ADDRESS (addr));
 	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	g_return_if_fail (priv->name || priv->sockaddr);
 
-	res_data = g_new (SoupAddressResolveAsyncData, 1);
-	res_data->addr          = addr;
+	res_data = g_slice_new0 (SoupAddressResolveAsyncData);
+	res_data->addr          = g_object_ref (addr);
 	res_data->callback      = callback;
 	res_data->callback_data = user_data;
 
-	g_object_ref (addr);
-	soup_dns_lookup_resolve_async (priv->lookup, async_context, cancellable,
-				       lookup_resolved, res_data);
+	if (priv->name && priv->sockaddr) {
+		res_data->status = SOUP_STATUS_OK;
+		soup_add_completion (async_context, idle_complete_resolve, res_data);
+		return;
+	}
+
+	resolver = g_resolver_get_default ();
+
+	if (async_context && async_context != g_main_context_default ())
+		g_main_context_push_thread_default (async_context);
+
+	if (priv->name) {
+		res_data->lookup_name = TRUE;
+		g_resolver_lookup_by_name_async (resolver, priv->name,
+						 cancellable,
+						 lookup_resolved, res_data);
+	} else {
+		GInetAddress *gia;
+
+		res_data->lookup_name = FALSE;
+		gia = soup_address_make_inet_address (addr);
+		g_resolver_lookup_by_address_async (resolver, gia,
+						    cancellable,
+						    lookup_resolved, res_data);
+		g_object_unref (gia);
+	}
+
+	if (async_context && async_context != g_main_context_default ())
+		g_main_context_pop_thread_default (async_context);
+
+	g_object_unref (resolver);
 }
 
 /**
@@ -618,15 +758,183 @@ guint
 soup_address_resolve_sync (SoupAddress *addr, GCancellable *cancellable)
 {
 	SoupAddressPrivate *priv;
+	GResolver *resolver;
+	GError *error = NULL;
 	guint status;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), SOUP_STATUS_MALFORMED);
 	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	g_return_val_if_fail (priv->name || priv->sockaddr, SOUP_STATUS_MALFORMED);
+
+	resolver = g_resolver_get_default ();
+
+	if (!priv->sockaddr) {
+		GList *addrs;
+
+		addrs = g_resolver_lookup_by_name (resolver, priv->name,
+						   cancellable, &error);
+		status = update_addrs (addr, addrs, error);
+		g_resolver_free_addresses (addrs);
+	} else if (!priv->name) {
+		GInetAddress *gia;
+		char *name;
+
+		gia = soup_address_make_inet_address (addr);
+		name = g_resolver_lookup_by_address (resolver, gia,
+						     cancellable, &error);
+		g_object_unref (gia);
+		status = update_name (addr, name, error);
+		g_free (name);
+	} else
+		status = SOUP_STATUS_OK;
+
+	if (error)
+		g_error_free (error);
+	g_object_unref (resolver);
 
-	g_object_ref (addr);
-	status = soup_dns_lookup_resolve (priv->lookup, cancellable);
-	if (status == SOUP_STATUS_OK)
-		update_address (addr, priv->lookup);
-	g_object_unref (addr);
 	return status;
 }
+
+/**
+ * soup_address_is_resolved:
+ * @addr: a #SoupAddress
+ *
+ * Tests if @addr has already been resolved.
+ *
+ * Return value: %TRUE if @addr has been resolved.
+ **/
+gboolean
+soup_address_is_resolved (SoupAddress *addr)
+{
+	SoupAddressPrivate *priv;
+
+	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), FALSE);
+	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+
+	return priv->sockaddr && priv->name;
+}
+
+/**
+ * soup_address_hash_by_name:
+ * @addr: a #SoupAddress
+ *
+ * A hash function (for #GHashTable) that corresponds to
+ * soup_address_equal_by_name(), qv
+ *
+ * Return value: the named-based hash value for @addr.
+ *
+ * Since: 2.26
+ **/
+guint
+soup_address_hash_by_name (gconstpointer addr)
+{
+	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+
+	g_return_val_if_fail (priv->name != NULL, 0);
+	return g_str_hash (priv->name);
+}
+
+/**
+ * soup_address_equal_by_name:
+ * @addr1: a #SoupAddress with a resolved name
+ * @addr2: another #SoupAddress with a resolved name
+ *
+ * Tests if @addr1 and @addr2 have the same "name". This method can be
+ * used with soup_address_hash_by_name() to create a #GHashTable that
+ * hashes on address "names".
+ *
+ * Comparing by name normally means comparing the addresses by their
+ * hostnames. But if the address was originally created using an IP
+ * address literal, then it will be compared by that instead.
+ *
+ * In particular, if "www.example.com" has the IP address 10.0.0.1,
+ * and @addr1 was created with the name "www.example.com" and @addr2
+ * was created with the name "10.0.0.1", then they will compare as
+ * unequal for purposes of soup_address_equal_by_name().
+ *
+ * This would be used to distinguish hosts in situations where
+ * different virtual hosts on the same IP address should be considered
+ * different. Eg, for purposes of HTTP authentication or cookies, two
+ * hosts with the same IP address but different names are considered
+ * to be different hosts.
+ *
+ * See also soup_address_equal_by_ip(), which compares by IP address
+ * rather than by name.
+ *
+ * Return value: whether or not @addr1 and @addr2 have the same name
+ *
+ * Since: 2.26
+ **/
+gboolean
+soup_address_equal_by_name (gconstpointer addr1, gconstpointer addr2)
+{
+	SoupAddressPrivate *priv1 = SOUP_ADDRESS_GET_PRIVATE (addr1);
+	SoupAddressPrivate *priv2 = SOUP_ADDRESS_GET_PRIVATE (addr2);
+
+	g_return_val_if_fail (priv1->name != NULL, FALSE);
+	g_return_val_if_fail (priv2->name != NULL, FALSE);
+	return !g_ascii_strcasecmp (priv1->name, priv2->name);
+}
+
+/**
+ * soup_address_hash_by_ip:
+ * @addr: a #SoupAddress
+ *
+ * A hash function (for #GHashTable) that corresponds to
+ * soup_address_equal_by_ip(), qv
+ *
+ * Return value: the IP-based hash value for @addr.
+ *
+ * Since: 2.26
+ **/
+guint
+soup_address_hash_by_ip (gconstpointer addr)
+{
+	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	guint hash;
+
+	g_return_val_if_fail (priv->sockaddr != NULL, 0);
+
+	memcpy (&hash, SOUP_ADDRESS_GET_DATA (priv),
+		MIN (sizeof (hash), SOUP_ADDRESS_FAMILY_DATA_SIZE (priv->sockaddr->sa_family)));
+	return hash;
+}
+
+/**
+ * soup_address_equal_by_ip:
+ * @addr1: a #SoupAddress with a resolved IP address
+ * @addr2: another #SoupAddress with a resolved IP address
+ *
+ * Tests if @addr1 and @addr2 have the same IP address. This method
+ * can be used with soup_address_hash_by_ip() to create a
+ * #GHashTable that hashes on IP address.
+ *
+ * This would be used to distinguish hosts in situations where
+ * different virtual hosts on the same IP address should be considered
+ * the same. Eg, if "www.example.com" and "www.example.net" have the
+ * same IP address, then a single #SoupConnection can be used to talk
+ * to either of them.
+ *
+ * See also soup_address_equal_by_name(), which compares by name
+ * rather than by IP address.
+ *
+ * Return value: whether or not @addr1 and @addr2 have the same IP
+ * address.
+ *
+ * Since: 2.26
+ **/
+gboolean
+soup_address_equal_by_ip (gconstpointer addr1, gconstpointer addr2)
+{
+	SoupAddressPrivate *priv1 = SOUP_ADDRESS_GET_PRIVATE (addr1);
+	SoupAddressPrivate *priv2 = SOUP_ADDRESS_GET_PRIVATE (addr2);
+	int size;
+
+	g_return_val_if_fail (priv1->sockaddr != NULL, FALSE);
+	g_return_val_if_fail (priv2->sockaddr != NULL, FALSE);
+
+	size = SOUP_ADDRESS_FAMILY_SOCKADDR_SIZE (priv1->sockaddr->sa_family);
+	return (priv1->sockaddr->sa_family ==
+		priv2->sockaddr->sa_family &&
+		!memcmp (priv1->sockaddr, priv2->sockaddr, size));
+}
